from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from .config import AwsConfig

    return AwsConfig


@plugins.register(plugins.AccountPlugin)
def account_types():
    from .accounts import AwsAccountContainer, KmsAccount

    AwsAccountContainer.model_rebuild()
    return AwsAccountContainer, KmsAccount



import pytest
from ape.types import MessageSignature, TransactionSignature
from ape_ethereum.transactions import StaticFeeTransaction
from eth_account.messages import encode_defunct

from ape_aws.accounts import AwsAccountContainer, KmsAccount


@pytest.fixture(scope="session")
def transaction():
    txn = {
        "chain_id": 11155111,
        "nonce": 0,
        "value": 1,
        "data": "0x00",
        "receiver": "0xa5D3241A1591061F2a4bB69CA0215F66520E67cf",
        "type": 0,
        "gas_limit": 1000000,
        "gas_price": 300000000000,
    }
    return StaticFeeTransaction(**txn)


@pytest.fixture(scope="session")
def string_message():
    return encode_defunct(text="12345")


@pytest.fixture(scope="session")
def aws_account_container():
    return AwsAccountContainer(name="ape-aws", account_type=KmsAccount)


@pytest.fixture(scope="session")
def kms_account(aws_account_container):
    for account in aws_account_container.accounts:
        if account.alias == "plugin-test":
            return account


def test_signing_message(kms_account, string_message):
    val = kms_account.sign_message(string_message)
    assert isinstance(val, MessageSignature)


def test_signing_transaction(kms_account, transaction):
    val = kms_account.sign_transaction(transaction)
    assert isinstance(val.signature, TransactionSignature)



import json
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, ClassVar

from ape.types import AddressType
from pydantic import BaseModel, Field, SecretStr, field_validator

from ape_aws.session import Session

if TYPE_CHECKING:
    from botocore.client import BaseClient  # type: ignore[import-untyped]


class ImportKey(BaseModel):
    private_key: SecretStr

    @property
    def address(self):
        from eth_account import Account

        return Account.privateKeyToAccount(self.private_key.get_secret()).address

    @property
    def ec_private_key(self):
        from cryptography.hazmat.primitives.asymmetric import ec

        return ec.derive_private_key(int(self.private_key, 16), ec.SECP256K1())

    @property
    def private_key_bin(self):
        """
        Returns the private key in binary format
        This is required for the `boto3.client.import_key_material` method
        """
        from cryptography.hazmat.primitives import serialization

        return self.ec_private_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    def public_key_der(self, import_token: bytes):
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization

        return serialization.load_der_public_key(import_token, backend=default_backend())

    def encrypt(self, import_token: bytes):
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import padding

        return self.public_key_der(import_token).encrypt(
            self.private_key_bin,
            padding.OAEP(
                mgf=padding.MGF1(hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )


class KmsKey(BaseModel):
    kms_client: ClassVar["BaseClient"]

    # NOTE: Using information from `aws kms list-aliases`
    id: str = Field(alias="TargetKeyId")
    arn: str = Field(alias="AliasArn")
    cached_alias: str | None = Field(default=None, alias="AliasName")
    cached_address: str | None = None

    @field_validator("cached_alias")
    def prune_alias_prepend(cls, value: str):
        return value.replace("alias/ape-aws/v1/", "")

    @property
    def alias(self) -> str:
        if self.cached_alias is not None:
            return self.cached_alias

        response = self.kms_client.list_aliases(KeyId=self.id)
        self.cached_alias = response["Aliases"][0]["AliasName"].replace("alias/ape-aws/v1/", "")
        assert isinstance(self.cached_alias, str)  # mypy
        return self.cached_alias

    @alias.setter  # type: ignore[attr-defined]
    def set_alias(self, alias: str):
        self.kms_client.update_alias(
            AliasName=f"alias/ape-aws/v1/{alias}",
            TargetKeyId=self.id,
            KeyUsage="SIGN_VERIFY",
        )
        self.cached_alias = alias

    @property
    def public_key(self):
        return self.kms_client.get_public_key(KeyId=self.id)["PublicKey"]

    @property
    def address(self) -> AddressType:
        if self.cached_address:
            return self.cached_address

        from eth_utils import keccak, to_checksum_address

        return AddressType(to_checksum_address(keccak(self.public_key[-64:])[-20:].hex().lower()))

    def add_tags(self, tags: list[dict[str, str]]):
        self.kms_client.tag_resource(KeyId=self.id, Tags=tags)

    def get_policy(self) -> dict:
        response = self.kms_client.get_key_policy(KeyId=self.id)
        return json.loads(response["Policy"])

    @staticmethod
    def USER_KEY_POLICY(user_arn: str, key_arn: str) -> dict:
        return dict(
            Sid="Allow use of a specific key",
            Effect="Allow",
            Principal=dict(AWS=user_arn),
            Action=["kms:Sign", "kms:Verify", "kms:GetPublicKey"],
            Resource=key_arn,
        )

    def set_policy(
        self,
        users_to_add: list[str] | None = None,
        users_to_remove: list[str] | None = None,
    ):
        policy = self.get_policy()

        for user_arn in users_to_remove or []:
            policy["Statement"] = [
                stmt for stmt in policy["Statement"] if stmt["Principal"]["AWS"] != user_arn
            ]

        for user_arn in users_to_add or []:
            for stmt in policy["Statement"]:
                if stmt["Principal"]["AWS"] == user_arn:
                    break

            else:
                policy["Statement"].append(self.USER_KEY_POLICY(user_arn, self.arn))

        self.kms_client.put_key_policy(
            KeyId=self.id,
            PolicyName="default",
            Policy=json.dumps(policy),
        )

    def sign(self, msghash: bytes) -> bytes:
        response = self.kms_client.sign(
            KeyId=self.id,
            Message=msghash,
            MessageType="DIGEST",
            SigningAlgorithm="ECDSA_SHA_256",
        )
        return response.get("Signature")

    def delete(self, days: int = 30):
        if self.alias:
            self.kms_client.delete_alias(AliasName=f"alias/ape-aws/v1/{self.alias}")

        self.kms_client.schedule_key_deletion(
            KeyId=self.id,
            PendingWindowInDays=days,
        )


class KmsClient(Session):
    @cached_property
    def kms_client(self):
        client = self.session.client("kms")
        KmsKey.kms_client = client  # DI
        return client

    @cached_property
    def keys(self) -> dict[str, KmsKey]:
        # NOTE: Uses aliases to get alias faster (no additional lookups needed)
        paginator = self.kms_client.get_paginator("list_aliases")
        pages = map(lambda data: data["Aliases"], paginator.paginate())
        # NOTE: Use `itertools.chain` since it is segmented into list of lists
        # NOTE: Just look for `alias/ape-aws/` alias in case we add v2, v3, etc.
        key_data = filter(lambda k: k["AliasName"].startswith("alias/ape-aws/"), chain(*pages))
        keys = map(KmsKey.model_validate, key_data)
        return {key.alias: key for key in keys}

    def generate_key(self, alias: str) -> KmsKey:
        response = self.kms_client.create_key(
            Description="Generated Key Created with Ape-AWS",
            KeyUsage="SIGN_VERIFY",
            KeySpec="ECC_SECG_P256K1",
            Origin="AWS_KMS",
        )

        key = KmsKey(
            TargetKeyId=response["KeyMetadata"]["KeyId"],
            AliasArn=response["KeyMetadata"]["Arn"],
        )

        self.kms_client.create_alias(
            TargetKeyId=key.id,
            AliasName=f"alias/ape-aws/v1{alias}",
        )

        key.cached_alias = alias
        self.keys.update({alias: key})  # Update cache
        return key

    def import_key(self, alias: str, private_key: ImportKey) -> KmsKey:
        response = self.kms_client.create_key(
            Description="Imported Key Created with Ape-AWS",
            KeyUsage="SIGN_VERIFY",
            KeySpec="ECC_SECG_P256K1",
            Origin="IMPORTED",
        )

        key = KmsKey(
            TargetKeyId=response["KeyMetadata"]["KeyId"],
            AliasArn=response["KeyMetadata"]["Arn"],
        )

        import_token_bytes = bytes.fromhex(response["ImportToken"])
        self.kms_client.import_key_material(
            KeyId=key.id,
            ImportToken=response["ImportToken"],
            EncryptedKeyMaterial=private_key.encrypt(import_token_bytes),
            ExpirationModel="KEY_MATERIAL_DOES_NOT_EXPIRE",
        )

        self.kms_client.create_alias(
            TargetKeyId=key.id,
            AliasName=f"alias/ape-aws/v1/{alias}",
            KeyUsage="SIGN_VERIFY",
        )

        key.cached_alias = alias
        self.keys.update({key.id: key})  # Update cache
        return key

    def delete_key(self, key_id: str, days: int = 30):
        key = self.keys.pop(key_id)  # Remove from cache
        key.delete(days=days)



import ecdsa  # type: ignore[import]


def _convert_der_to_rsv(signature: bytes, v_adjustment_factor: int = 0) -> dict:
    r, s = ecdsa.util.sigdecode_der(signature, ecdsa.SECP256k1.order)
    if s > ecdsa.SECP256k1.order / 2:
        s = ecdsa.SECP256k1.order - s

    return dict(
        r=r.to_bytes(32, byteorder="big"),
        s=s.to_bytes(32, byteorder="big"),
        v=v_adjustment_factor,
    )



from functools import cached_property
from typing import Any, Iterator, Optional

from ape.api.accounts import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, SignableMessage, TransactionSignature
from eth_account._utils.legacy_transactions import serializable_unsigned_transaction_from_dict
from eth_account.messages import _hash_eip191_message, encode_defunct
from eth_pydantic_types import HexBytes
from eth_typing import Hash32

from ape_aws.exceptions import ApeAwsException

from .client import AwsClient
from .kms import KmsKey
from .utils import _convert_der_to_rsv


class AwsAccountContainer(AwsClient, AccountContainerAPI):
    def __init__(self, *args, **kwargs):
        super(AwsClient, self).__init__()  # NOTE: Use config/envvar default
        super(AccountContainerAPI, self).__init__(*args, **kwargs)

    @property
    def aliases(self) -> Iterator[str]:
        yield from iter(self.keys)

    def __len__(self) -> int:
        return len(self.keys)

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        return map(lambda key: KmsAccount(key=key), self.keys.values())


class KmsAccount(AccountAPI):
    key: KmsKey

    @property
    def alias(self) -> str:
        return self.key.alias

    @cached_property
    def address(self) -> AddressType:
        return self.key.address

    def sign_raw_msghash(self, msghash: HexBytes | Hash32) -> Optional[MessageSignature]:
        if len(msghash) != 32:
            return None

        if not (signature := self.key.sign(msghash)):
            return None

        msg_sig = MessageSignature(**_convert_der_to_rsv(signature, 27))
        if not self.check_signature(msghash, msg_sig, recover_using_eip191=False):
            # NOTE: Best way to determine parity is just to try recovery
            msg_sig = MessageSignature(v=msg_sig.v + 1, r=msg_sig.r, s=msg_sig.s)

        return msg_sig

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        if isinstance(msg, SignableMessage):
            message = msg
        elif isinstance(msg, str):
            if msg.startswith("0x"):
                message = encode_defunct(hexstr=msg)
            else:
                message = encode_defunct(text=msg)
        elif isinstance(msg, bytes):
            message = encode_defunct(primitive=msg)
        else:
            raise ApeAwsException(f"Cannot sign {type(msg)}")

        return self.sign_raw_msghash(_hash_eip191_message(message))

    def sign_transaction(self, txn: TransactionAPI, **signer_options) -> Optional[TransactionAPI]:
        """
        Sign an EIP-155 transaction.
        CHECK TYPE 0 transactions.

        Args:
            txn (``TransactionAPI``): A pydantic model of transaction data.

        Returns:
            TransactionAPI | None
        """
        txn_dict = txn.model_dump()
        # NOTE: remove this so doesn't raise below
        txn_dict.pop("from", None)
        unsigned_txn_hash = serializable_unsigned_transaction_from_dict(txn_dict).hash()

        if not (signature := self.sign_raw_msghash(HexBytes(unsigned_txn_hash))):
            return None

        # NOTE: We already added 27 to v above, so substract it from v to normalize to {0,1}
        v = signature.v - 27
        if txn.chain_id and txn.type == 0:
            # Include 155 chain ID protection offset
            v += (2 * txn.chain_id) + 35

        txn.signature = TransactionSignature(v=v, r=signature.r, s=signature.s)
        return txn



from typing import TYPE_CHECKING

import click

from ape_aws.exceptions import ApeAwsException

from .iam import IamClient
from .kms import KmsClient
from .session import Session

if TYPE_CHECKING:
    from click import Context


# NOTE: Full AWS Client is far too complicated to try to include in one class
class AwsClient(IamClient, KmsClient, Session):
    """The fully assembled AWS Client"""

    @classmethod
    def _click_argument_callback(cls, ctx: "Context", arg: str, value: str | None) -> "AwsClient":
        try:
            return cls(value)
        except ApeAwsException as e:
            raise click.BadOptionUsage("profile", str(e))



import json
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, ClassVar

from pydantic import BaseModel, Field, model_validator

from ape_aws.session import Session

if TYPE_CHECKING:
    from botocore.client import BaseClient  # type: ignore[import-untyped]


class IamPolicy(BaseModel):
    arn: str = Field(alias="PolicyArn")
    name: str = Field(alias="PolicyName")

    @model_validator(mode="before")
    @classmethod
    def patch_policy(cls, p: dict) -> dict:
        if "Arn" in p and "PolicyArn" not in p:
            p["PolicyArn"] = p.pop("Arn")

        return p


class IamAccessKey(BaseModel):
    id: str = Field(alias="AccessKeyId")
    status: str = Field(alias="Status")
    creation: datetime = Field(alias="CreateDate")
    secret_key: str | None = Field(default=None, alias="SecretAccessKey")

    def to_environment(self) -> str:
        if not self.secret_key:
            raise ValueError("Cannot view access key after creation, please delete and try again.")
        # NOTE: This matches up with loading code from `.session.Session.__init__`
        return f"AWS_ACCOUNT_ID={self.id}\nAWS_SECRET_KEY={self.secret_key}"


class IamUser(BaseModel):
    iam_client: ClassVar["BaseClient"]

    id: str = Field(alias="UserId")
    arn: str = Field(alias="Arn")
    name: str = Field(alias="UserName")
    creation: datetime = Field(alias="CreateDate")

    @cached_property
    def policies(self) -> dict[str, IamPolicy]:
        result = self.iam_client.list_attached_user_policies(UserName=self.name)
        policies = map(IamPolicy.model_validate, result["AttachedPolicies"])
        return {policy.name: policy for policy in policies}

    def add_policy(self, policy: IamPolicy):
        self.iam_client.attach_user_policy(UserName=self.name, PolicyArn=policy.arn)
        self.policies.update({policy.name: policy})  # update cache

    def remove_policy(self, policy: IamPolicy):
        self.iam_client.detach_user_policy(UserName=self.name, PolicyArn=policy.arn)
        del self.policies[policy.name]  # update cache

    @property
    def is_admin(self) -> bool:
        return self.policies.get("AdministratorAccess") is not None

    @cached_property
    def access_keys(self) -> dict[str, IamAccessKey]:
        response = self.iam_client.list_access_keys(UserName=self.name)
        access_keys = map(IamAccessKey.model_validate, response["AccessKeyMetadata"])
        return {key.id: key for key in access_keys}

    def create_access_key(self) -> IamAccessKey:
        response = self.iam_client.create_access_key(UserName=self.name)
        key = IamAccessKey.model_validate(response["AccessKey"])
        self.access_keys.update({key.id: key})  # Update cache
        return key

    def delete_access_key(self, key_id: str):
        key = self.access_keys.pop(key_id)  # Remove from cache
        self.iam_client.delete_access_key(UserName=self.name, AccessKeyId=key.id)

    def delete(self):
        while len(self.access_keys) > 0:
            self.delete_access_key(list(self.access_keys)[0])

        while len(self.policies) > 0:
            self.remove_policy(list(self.policies.values())[0])

        self.iam_client.delete_user(UserName=self.name)


class IamClient(Session):
    """AWS Client API for working with IAM Accounts"""

    KEY_POLICY_NAME = "ApeAwsKeyAccessV1"
    KEY_ACCESS_POLICY = dict(
        Sid="ApeAWSv1",
        Effect="Allow",
        Action=["kms:ListAliases", "kms:Sign", "kms:Verify", "kms:GetPublicKey"],
        Resource="*",
    )

    @cached_property
    def iam_client(self):
        client = self.session.client("iam")
        IamUser.iam_client = client  # DI
        return client

    @cached_property
    def users(self) -> dict[str, IamUser]:
        response = self.iam_client.list_users()
        users = map(IamUser.model_validate, response.get("Users", []))
        return {user.name: user for user in users}

    def create_user(self, user_name: str) -> IamUser:
        result = self.iam_client.create_user(UserName=user_name)
        user = IamUser.model_validate(result["User"])
        self.users.update({user.name: user})  # Update cache
        return user

    def delete_user(self, user_name: str):
        user = self.users.pop(user_name)  # Remove from cache
        user.delete()

    @cached_property
    def policies(self) -> dict[str, IamPolicy]:
        result = self.iam_client.list_policies()
        policies = map(IamPolicy.model_validate, result["Policies"])
        return {policy.name: policy for policy in policies}

    def create_key_policy(self) -> IamPolicy:
        response = self.iam_client.create_policy(
            PolicyName=self.KEY_POLICY_NAME,
            PolicyDocument=json.dumps(
                dict(
                    Version="2012-10-17",
                    Statement=[self.KEY_ACCESS_POLICY],
                )
            ),
        )

        policy = IamPolicy.model_validate(response["Policy"])
        self.policies.update({policy.name: policy})  # Update cache
        return policy



import os
from typing import TYPE_CHECKING, ClassVar

from ape.utils import ManagerAccessMixin
from boto3 import Session as AwsSession  # type: ignore[import-untyped]
from botocore.exceptions import NoCredentialsError, ProfileNotFound  # type: ignore[import-untyped]

from ape_aws.exceptions import ApeAwsException

if TYPE_CHECKING:
    from .config import AwsConfig


# NOTE: Subclass this w/ other clients to access `.session`
class Session(ManagerAccessMixin):
    """Helper for loading AWS Session profile w/ Ape Config"""

    session: ClassVar[AwsSession]

    @property
    def config(self) -> "AwsConfig":
        return self.config_manager.get_config("aws")

    def __init__(self, profile_name: str | None = None):
        session_kwargs: dict[str, str] = {}

        if profile_name is not None:
            session_kwargs["profile_name"] = profile_name

        elif (account_id := os.environ.get("AWS_ACCOUNT_ID")) and (
            secret_key := os.environ.get("AWS_SECRET_KEY")
        ):
            session_kwargs["aws_access_key_id"] = account_id
            session_kwargs["aws_secret_access_key"] = secret_key

        elif profile_envvar := os.environ.get("AWS_PROFILE"):
            session_kwargs["profile_name"] = profile_envvar

        elif self.config.default_profile:
            session_kwargs["profile_name"] = self.config.default_profile

        else:
            session_kwargs["profile_name"] = "default"

        if region_name := os.environ.get("AWS_DEFAULT_REGION"):
            session_kwargs["region_name"] = region_name

        elif self.config.default_region:
            session_kwargs["region_name"] = self.config.default_region

        else:
            session_kwargs["region_name"] = "us-east-1"

        try:
            self.__class__.session = AwsSession(**session_kwargs)

        except (NoCredentialsError, ProfileNotFound) as e:
            raise ApeAwsException(str(e)) from e



from collections import defaultdict

import click

from ape_aws.client import AwsClient


def _click_profile_option(argument_name):

    return click.option(
        "-p",
        "--profile",
        argument_name,
        default=None,
        callback=AwsClient._click_argument_callback,
        help=(
            "AWS profile to use to run this command "
            "(set default via Ape config, or use `AWS_PROFILE_ID`)"
        ),
    )


@click.group()
def cli():
    """
    Configure AWS Accounts and Signing Keys for use with Ape

    This command helps you to create AWS IAM accounts with the proper permissions to use as a
    remote signer, as well as create keys in AWS KMS that those accounts can access to sign with
    via Ape. The goal of this plugin is to configure a cloud environment to use for automation,
    such as running on the Silverback Platform or other production use cases.

    To begin, start with `ape aws whoami` to make sure your AWS account is properly loaded.
    """


@cli.command()
@_click_profile_option("client")
def whoami(client: AwsClient):
    """Display profile information of logged-in account"""
    click.echo(f"Region: {client.session.region_name}")
    click.echo(f"Profile: {client.session.profile_name}")
    creds = client.session.get_credentials()
    click.echo(f"Access Key: {creds.access_key}")


@cli.command()
@_click_profile_option("client")
def policies(client: AwsClient):
    """List all available policies

    Note that if you want to create a new policy, it is recommended to do so through AWS console,
    or the `aws` cli.
    """
    if policies := "\n".join(
        f"{policy.name} ({policy.arn})" for policy in client.policies.values()
    ):
        click.echo(policies)


@cli.group()
def users():
    """Commands for configuring IAM Users, Policies, and more"""


@users.command(name="list")
@_click_profile_option("client")
def list_user(client: AwsClient):
    """Display available users accessible with the logged in profile"""
    if users := "\n".join(f"{user.name} ({user.arn})" for user in client.users.values()):
        click.echo(users)

    else:
        raise click.UsageError(
            "No users available under profile '{aws_client.profile}'.\n\n"
            "Try `ape aws add-user` to create a new signer, or switch profiles."
        )


@users.command(name="new")
@_click_profile_option("client")
@click.option("-p", "--policy", "policy_names", multiple=True, help="Name of a valid policy")
@click.argument("name")
def add_user(client: AwsClient, policy_names: list[str], name: str):
    """Add a new IAM user NAME under the logged in profile

    Note that you can also add policies to this IAM user through the `--policy` option
    """
    user = client.create_user(name)
    click.echo(f"Created {user.name} ({user.arn})")

    if client.KEY_POLICY_NAME not in client.policies:
        key_policy = client.create_key_policy()
        click.echo(f"Policy '{key_policy.name}' added.")
        policy_names = [*policy_names, key_policy.name]

    elif client.KEY_POLICY_NAME not in policy_names:
        policy_names = [*policy_names, client.KEY_POLICY_NAME]

    for policy_name in policy_names:
        if not (policy := client.policies.get(policy_name)):
            click.echo(f"'{policy_name}' is not a valid policy name", err=True)
            continue

        user.add_policy(policy)
        click.echo(f"Policy '{policy.name}' added to user '{user.name}'.")


@users.command(name="remove")
@_click_profile_option("client")
@click.argument("name")
def remove_user(client: AwsClient, name: str):
    """Remove IAM user NAME under the logged in profile"""
    if not (user := client.users.get(name)):
        raise click.UsageError(f"'{name}' is not an available IAM Account")

    user.delete()
    click.echo(f"Removed {user.name} ({user.arn})")


@users.group()
def tokens():
    """Commands for configuring access tokens for an IAM user"""


@tokens.command(name="list")
@_click_profile_option("client")
@click.argument("name")
def list_access_tokens(client: AwsClient, name: str):
    """List access tokens for the IAM user NAME"""
    if not (user := client.users.get(name)):
        raise click.UsageError(f"'{name}' is not an available IAM Account")

    if keys := "\n".join(
        f"{key.id} ({key.status}) - Created: {key.creation}" for key in user.access_keys.values()
    ):
        click.echo(keys)


@tokens.command(name="new")
@_click_profile_option("client")
@click.argument("name")
def new_access_token(client: AwsClient, name: str):
    """Create an access token for the IAM user NAME

    Note that you cannot view this key after it is generated, so please store it carefully.
    Also note that this access key gives full access to the IAM account it is generated for.

    You can use this key to automatically grant access to the IAM account it is generated for
    via the environment variables mentioned when called (e.g. in a containerized environment).
    """
    if not (user := client.users.get(name)):
        raise click.UsageError(f"'{name}' is not an available IAM Account")

    access_key = user.create_access_key()

    # NOTE: Do these to `err=True` so you can do `ape aws users access-key new ALIAS > .env`
    click.secho(f"SUCCESS: Access key created for '{name}'", fg="green", err=True)
    click.secho(
        "WARNING: Access key will not be available after this command", fg="bright_yellow", err=True
    )
    click.echo(access_key.to_environment())


@tokens.command(name="remove")
@_click_profile_option("client")
@click.argument("name")
@click.argument("key_id")
def remove_access_token(client: AwsClient, name: str, key_id: str):
    """Remove access token KEY_ID for the IAM user NAME"""
    if not (user := client.users.get(name)):
        raise click.UsageError(f"'{name}' is not an available IAM Account")

    if not user.access_keys.get(key_id):
        raise click.UsageError(f"'{key_id}' is not an available Access Key for {user.name}")

    user.delete_access_key(key_id)
    click.secho(f"Access key {key_id} removed from {user.name}", fg="green")


@cli.group()
def keys():
    """Commands for creating and managing KMS keys, access controls, and more

    It is recommended that you use `ape aws users` to create a new IAM account with the
    correct attached policies needed to use KMS and perform key signing via Ape.
    """


@keys.command(name="list")
@_click_profile_option("client")
def list_keys(client: AwsClient):
    """List all keys available to account"""
    if keys := "\n".join(f"'{key.alias}' (id: {key.id})" for key in client.keys.values()):
        click.echo(keys)


@keys.command()
@_click_profile_option("client")
@click.option(
    "-u",
    "--user",
    "user_names",
    multiple=True,
    help="Give access to key for one or more users, ex. -u NAME1, -u NAME2",
)
@click.argument("alias")
def generate(client: AwsClient, user_names: list[str], alias: str):
    """Generate a new key ALIAS and grant access to USERS and ADMINS"""
    key = client.generate_key(alias)
    click.echo(f"Key created '{key.id}'.")

    users = []
    for user_name in user_names:
        if not (user := client.users.get(user_name)):
            click.echo(f"'{user_name}' is not a valid user name for an admin", err=True)
            continue

        users.append(user)

    key.set_policy(users_to_add=[u.arn for u in users])
    click.secho(f"Key policies updated for {key.alias}", fg="green", err=True)


@keys.command(name="show")
@_click_profile_option("client")
@click.argument("alias")
def show_key_info(client: AwsClient, alias: str):
    """Show info about key ALIAS, including Admins and Users"""
    if not (key := client.keys.get(alias)):
        raise click.UsageError(f"'{alias}' is not an available key alias")

    key_policy = key.get_policy()
    access_rights: dict[str, dict[str, str | list[str]]] = defaultdict(dict)
    for stmt in key_policy["Statement"]:
        if (arn := stmt["Principal"]["AWS"]).endswith("root"):
            continue

        for user in client.users.values():
            if user.arn == arn:
                access_rights[arn][stmt["Resource"]] = stmt["Action"]
                break

        else:
            click.echo(f"Unrecognized ARN: {arn}", err=True)

    for user_arn, key_access in access_rights.items():
        click.echo(f"{user_arn}:")
        for key_arn, rights in key_access.items():
            if isinstance(rights, list):
                rights = ", ".join(rights)
            click.echo(f"  {key_arn}: {rights}")


@keys.command(name="grant")
@_click_profile_option("client")
@click.option(
    "-u",
    "--user",
    "user_names",
    multiple=True,
    help="Give access to key for one or more users, ex. -u NAME1, -u NAME2",
)
@click.argument("alias")
def grant_key_access(client: AwsClient, user_names: list[str], alias: str):
    """Grant access to USERS and ADMINS for key ALIAS"""
    if not (key := client.keys.get(alias)):
        raise click.UsageError(f"'{alias}' is not an available key alias")

    users = []
    for user_name in user_names:
        if not (user := client.users.get(user_name)):
            click.echo(f"'{user_name}' is not a valid user name for an admin", err=True)
            continue

        users.append(user)

    key.set_policy(users_to_add=[u.arn for u in users])
    click.secho(f"Key policies updated for {key.alias}", fg="green", err=True)


@keys.command(name="revoke")
@_click_profile_option("client")
@click.option(
    "-u",
    "--user",
    "user_names",
    multiple=True,
    help="Give access to key for one or more users, ex. -u NAME1, -u NAME2",
)
@click.argument("alias")
def revoke_key_access(client: AwsClient, user_names: list[str], alias: str):
    """Revoke access to USERS and ADMINS for key ALIAS"""
    if not (key := client.keys.get(alias)):
        raise click.UsageError(f"'{alias}' is not an available key alias")

    users = []
    for user_name in user_names:
        if not (user := client.users.get(user_name)):
            click.echo(f"'{user_name}' is not a valid user name for an admin", err=True)
            continue

        users.append(user)

    key.set_policy(users_to_remove=[u.arn for u in users])
    click.secho(f"Key policies updated for {key.alias}", fg="green", err=True)
